import json
from urllib.parse import urljoin

# Used for the synchronous session
import requests

# Used for the asynchronous session
import aiohttp

import pprint
from typing import Dict

DEFAULT_URL = "http://localhost:1324"


class HeedyError(Exception):
    """
    HeedyError is raised when the server returns an error value
    in response to a request.

    You can get the error contents by accessing the :code:"error" property
    and the :code:"error_description" property::

        # This is equivalent to print(myerror)
        print(f"{myerror.error}: {myerror.error_description}")
    """

    def __init__(self, msg):
        self.error = msg["error"]
        self.error_description = msg["error_description"]

    def __str__(self):
        return f"{self.error}: {self.error_description}"


class Session:
    """
    Session is the abstract base class that both sync and async sessions implement
    """

    def __init__(self, namespace, url=DEFAULT_URL):
        self.namespace = namespace

        # Set up the API url
        if not url.startswith("http"):
            url = "https://" + url
        if not url.endswith("/"):
            url = url + "/"
        self.url = url

    def f(self, func, x):
        raise NotImplementedError()

    @property
    def isasync(self):
        raise NotImplementedError()

    def setAccessToken(self, token):
        raise NotImplementedError()

    def setPluginKey(self, key):
        raise NotImplementedError()

    def setHeader(self, key, value):
        raise NotImplementedError()

    def version(self):
        raise NotImplementedError()

    def get(self, path, params={}, f=lambda x: x):
        raise NotImplementedError()

    def post(self, path, data, params={}, f=lambda x: x):
        raise NotImplementedError()

    def patch(self, path, data, params={}, f=lambda x: x):
        raise NotImplementedError()

    def delete(self, path, params={}, f=lambda x: x):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()


class SyncSession(Session):
    """
    SyncSession is to be used in synchronous programs. It uses requests internally.
    """

    def __init__(self, namespace, url=DEFAULT_URL):
        super().__init__(namespace, url)
        self.s = requests.Session()
        self.s.headers.update({"Content-Type": "application/json"})

    def f(self, x, func):
        return func(x)

    @property
    def isasync(self):
        return True

    def setAccessToken(self, token):
        self.s.headers.update({"Authorization": f"Bearer {token}"})

    def setPluginKey(self, key):
        self.s.headers.update({"X-Heedy-Key": key})

    def setHeader(self, key, value):
        self.s.headers.update({key: value})

    def version(self):
        return self.handleResponse(
            self.s.get(urljoin(self.url, "api/server/version"))
        ).text

    def handleResponse(self, r):
        if r.status_code >= 400:
            # The response returned an error
            try:
                msg = r.json()
            except:
                msg = {
                    "error": "malformed_response",
                    "error_description": f'The server returned "{r.text}", which is not json.',
                }
            raise HeedyError(msg)
        return r

    def get(self, path, params={}, f=lambda x: x):
        return f(
            self.handleResponse(
                self.s.get(urljoin(self.url, path), params=params)
            ).json()
        )

    def post(self, path, data, params={}, f=lambda x: x):
        return f(
            self.handleResponse(
                self.s.post(
                    urljoin(self.url, path), data=json.dumps(data), params=params
                )
            ).json()
        )

    def patch(self, path, data, params={}, f=lambda x: x):
        return f(
            self.handleResponse(
                self.s.patch(
                    urljoin(self.url, path), data=json.dumps(data), params=params
                )
            ).json()
        )

    def delete(self, path, params={}, f=lambda x: x):
        return f(
            self.handleResponse(
                self.s.delete(urljoin(self.url, path), params=params)
            ).json()
        )

    def close(self):
        self.s.close()


class AsyncSession(Session):
    """
    AsyncSession is used when running in an asyncio event loop. All of the requests become coroutines,
    allowing them to be awaited
    """

    def __init__(self, namespace, url=DEFAULT_URL):
        super().__init__(namespace, url)
        self.s = None
        self.headers = {"Content-Type": "application/json"}

    @property
    def isasync(self):
        return True

    async def f(self, x, func):
        return func(await x)

    @staticmethod
    def __p(p):
        for k in p:
            if not isinstance(p[k], str):
                p[k] = json.dumps(p[k])
        return p

    def setAccessToken(self, token):
        self.headers["Authorization"] = f"Bearer {token}"

    def setPluginKey(self, key):
        self.headers["X-Heedy-Key"] = key

    def setHeader(self, key, value):
        self.headers[key] = value

    def initSession(self):
        if self.s is None:
            self.s = aiohttp.ClientSession()

    async def handleResponse(self, r):
        if r.status >= 400:
            # The response returned an error
            try:
                msg = await r.json()
            except:
                msg = {
                    "error": "malformed_response",
                    "error_description": f"The server did not return valid json",
                }
            raise HeedyError(msg)
        return r

    async def version(self):
        self.initSession()
        return await (
            await self.handleResponse(
                await self.s.get(urljoin(self.url, "api/server/version"))
            )
        ).text()

    async def get(self, path, params={}, f=lambda x: x):
        self.initSession()
        return f(
            await (
                await self.handleResponse(
                    await self.s.get(
                        urljoin(self.url, path),
                        params=self.__p(params),
                        headers=self.headers,
                    )
                )
            ).json()
        )

    async def post(self, path, data, params={}, f=lambda x: x):
        self.initSession()
        return f(
            await (
                await self.handleResponse(
                    await self.s.post(
                        urljoin(self.url, path),
                        params=self.__p(params),
                        data=json.dumps(data),
                        headers=self.headers,
                    )
                )
            ).json()
        )

    async def patch(self, path, data, params={}, f=lambda x: x):
        self.initSession()
        return f(
            await (
                await self.handleResponse(
                    await self.s.patch(
                        urljoin(self.url, path),
                        params=self.__p(params),
                        data=json.dumps(data),
                        headers=self.headers,
                    )
                )
            ).json()
        )

    async def delete(self, path, params={}, f=lambda x: x):
        self.initSession()
        return f(
            await (
                await self.handleResponse(
                    await self.s.delete(
                        urljoin(self.url, path),
                        params=self.__p(params),
                        headers=self.headers,
                    )
                )
            ).json()
        )

    async def raw(self, method, path, data=None, params={}, headers={}):
        self.initSession()
        return await self.s.request(
            method,
            urljoin(self.url, path),
            headers={**self.headers, **headers},
            data=data,
        )

    async def close(self):
        if self.s is not None:
            await self.s.close()


def getSessionType(sessionType: str, namespace: str, url: str = DEFAULT_URL) -> Session:
    """
    This function is given a string, either "sync" or "async", and it returns a SyncSession or AsyncSession respectively.
    """
    if sessionType == "sync":
        return SyncSession(namespace, url)
    if sessionType == "async":
        return AsyncSession(namespace, url)
    raise NotImplementedError(
        f"The session type '{sessionType}' is not implemented")

from .notifications import Notifications

class APIObject:
    """
    APIObject represents an object in heedy (user,app,object,etc).
    It is given a session and the api location of the object, and allows
    reading, updating, and deleting the object
    """

    props = {"name", "description", "icon"}

    def __init__(self, uri: str, constraints: Dict, session: Session, cached_data = {}):
        self.session = session
        self.uri = uri
        self.cached_data = cached_data

        self.notifications = Notifications(constraints, self.session)

    def read(self, **kwargs):
        """
        Read the object
        """
        def writeCache(o):
            self.cached_data = o
            return o
        return self.session.f(self.session.get(self.uri, params=kwargs),writeCache)

    def update(self, **kwargs):
        """
        Updates the given data::

            o.update(name="My new name",description="my new description")
        """
        return self.session.patch(self.uri, kwargs)

    def delete(self, **kwargs):
        """
        Deletes the object
        """
        return self.session.delete(self.uri, params=kwargs)

    def __setattr__(self, name, value):
        if name in self.props:
            return self.update(**{name: value})
        return super().__setattr__(name, value)

    def __getattr__(self, attr: str):
        return self.session.f(self.read(), lambda x: x[attr])

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return other.uri == self.uri
        return False

    def __getitem__(self,i):
        # Gets the item from the cache - assumes that the data is in the cache. If not, need to call .read() first
        return self.cached_data[i]

    def __str__(self):
        return self.__class__.__name__ + pprint.pformat(self.cached_data)

    def __repr__(self):
        return str(self)

    def notify(self, *args, **kwargs):
        return self.notifications.notify(*args, **kwargs)


class APIList:
    """
    APIList represents a list of objects in heedy (users,apps,objects,etc).
    """

    def __init__(self, uri: str, constraints: Dict, session: Session):
        self.session = session
        self.uri = uri
        self._constraints = constraints

    # These are internal functions that help with implementing the useful parts of
    # lists
    def _create(self, f=lambda x: x, **kwargs):
        return self.session.post(self.uri, {**self._constraints, **kwargs}, f=f)

    def _getitem(self, item, f=lambda x: x):
        return self.session.get(f"{self.uri}/{item}", f=f)

    def _call(self, f=lambda x: x, **kwargs):
        return self.session.get(self.uri, params={**self._constraints, **kwargs}, f=f)



from typing import Dict

from ..base import APIObject, APIList, Session
from ..kv import KV

from .. import users
from .. import apps
from ..notifications import Notifications

from . import registry


class Object(APIObject):
    props = {"name", "description", "icon", "meta"}

    def __init__(self, objectData: Dict, session: Session):
        super().__init__(
            f"api/objects/{objectData['id']}",
            {"object": objectData["id"]},
            session,
            cached_data=objectData
        )
        self._kv = KV(f"api/kv/objects/{objectData['id']}", self.session)

    @property
    def kv(self):
        return self._kv

    @kv.setter
    def kv(self, v):
        return self._kv.set(**v)

    def __getattr__(self, attr):
        return self.cached_data[attr]

    @property
    def owner(self):
        return users.User(self.cached_data["owner"], self.session)

    @property
    def app(self):
        if self.cached_data["app"] is None:
            return None
        return apps.App(self.cached_data["app"], session=self.session)

class Objects(APIList):
    def __init__(self, constraints: Dict, session: Session):
        super().__init__("api/objects", constraints, session)

    def __getitem__(self, item):
        return super()._getitem(item, f=lambda x: registry.getObject(x, self.session))

    def __call__(self, **kwargs):
        return super()._call(
            f=lambda x: [registry.getObject(xx, self.session) for xx in x], **kwargs
        )

    def create(self, name, meta={}, otype="timeseries", **kwargs):
        """
        Creates a new object of the given type (timeseries by default).
        """
        return super()._create(
            f=lambda x: registry.getObject(x, self.session),
            **{"name": name, "type": otype, "meta": meta, **kwargs},
        )




from .base import Session
import urllib.parse


class KV:
    def __init__(self, uri: str, session: Session):
        self.session = session
        self.uri = uri

    def _uri(self, namespace):
        if namespace is None:
            namespace = self.session.namespace
        return self.uri + "/" + namespace

    def get(self, namespace=None):
        return self.session.get(self._uri(namespace))

    def getkey(self, key: str, namespace=None):
        return self.session.get(
            self._uri(namespace) + "/" + urllib.parse.quote(key, safe="")
        )

    def set(self, namespace=None, **kwargs):
        return self.session.post(self._uri(namespace), kwargs)

    def update(self, namespace=None, **kwargs):
        return self.session.patch(self._uri(namespace), kwargs)

    def delete(self, key, namespace=None):
        return self.session.delete(
            self._uri(namespace) + "/" + urllib.parse.quote(key, safe="")
        )

    def __getitem__(self, key: str):
        return self.getkey(key)

    def __setitem__(self, key: str, value):
        return self.set(**{key: value})

    def __delitem__(self, key: str):
        return self.delete(key)

    def __call__(self, namespace=None):
        return self.get(namespace)



from typing import Dict
from .base import APIObject, APIList, Session, getSessionType, DEFAULT_URL
from .kv import KV

from . import users
from . import objects
from .notifications import Notifications

from functools import partial


class App(APIObject):
    props = {"name", "description", "icon", "settings", "settings_schema"}

    def __init__(self, access_token: str, url: str = DEFAULT_URL, session="sync", cached_data={}):
        appid = "self"
        if isinstance(session, Session):
            # Treat the session as already initialized, meaning that the access token is actually
            # the app id
            appid = access_token
            super().__init__(f"api/apps/{appid}", {"app": appid}, session, cached_data=cached_data)

        else:
            # Initialize the app object as a direct API
            s = getSessionType(session, "self")
            s.setAccessToken(access_token)
            super().__init__("api/apps/self", {"app": appid}, s)
        # The objects belonging to the app
        self.objects = objects.Objects({"app": appid}, self.session)
        # Key-value store associated with the app
        self.kv = KV(f"api/kv/apps/{appid}", self.session)

    @property
    def owner(self):
        return self.session.f(
            self.read(), lambda x: users.User(x["owner"], self.session)
        )


class Apps(APIList):
    def __init__(self, constraints: Dict, session: Session):
        super().__init__("api/apps", constraints, session)

    def __getitem__(self, item):
        return self._getitem(item, f=lambda x: App(x["id"], session=self.session,cached_data=x))

    def __call__(self, **kwargs):
        return self._call(
            f=lambda x: [App(xx["id"], session=self.session,cached_data=xx) for xx in x], **kwargs
        )

    def create(self, name, otype="timeseries", **kwargs):
        return self._create(
            f=lambda x: App(x["id"], session=self.session,cached_data=x),
            **{"name": name, "type": otype, **kwargs},
        )



from .objects import Object
from .registry import registerObjectType

import time
import datetime
import json
import os.path
from urllib.parse import urljoin

# Allows querying by string times
from dateparser import parse


def parseTime(t):
    if isinstance(t, str):
        t = parse(t)
        if t is None:
            raise AttributeError("Could not parse timestamp")
    if isinstance(t, datetime.datetime):
        t = t.timestamp()
    return t


class DatapointArray(list):
    """
    The DatapointArray is a convenience wrapper on data returned from timeseries.
    It allows a bit of extra functionality to make working with timeseries simpler.
    """

    def __init__(self, data=[]):
        list.__init__(self, data)

    def __add__(self, other):
        return DatapointArray(self).merge(other)

    def __radd__(self, other):
        return DatapointArray(self).merge(other)

    def __getitem__(self, key):
        if key == "t":
            return self.t()
        if key == "d":
            return self.d()
        d = list.__getitem__(self, key)
        if isinstance(key, slice):
            d = DatapointArray(d)
            d._d = self._d[key]
            d._t = self._t[key]
        return d

    def sort(self, f=lambda d: d["t"]):
        """Sort here works by sorting by timestamp by default"""
        list.sort(self, key=f)
        return self

    def d(self):
        """Returns just the data portion of the datapoints as a list"""
        return list(map(lambda x: x["d"], self.raw()))

    def t(self):
        """Returns just the timestamp portion of the datapoints as a list.
        The timestamps are in python datetime's date format."""
        return list(map(lambda x: datetime.datetime.fromtimestamp(x["t"]), self.raw()))

    def dt(self):
        """Returns just the durations of all datapoints.
        """
        return list(map(lambda x: 0.0 if not "dt" in x else x["dt"], self.raw()))

    def merge(self, array):
        """Adds the given array of datapoints to the generator.
        It assumes that the datapoints are formatted correctly for heedy, meaning
        that they are in the format::

            [{"t": unix timestamp, "d": data}]

        The data does NOT need to be sorted by timestamp - this function sorts it for you
        """
        self.extend(array)
        self.sort()

        return self

    def raw(self):
        """Returns array as a raw python array. For cases where for some reason
        the DatapointArray wrapper does not work for you

        """
        return list.__getitem__(self, slice(None, None))

    def writeJSON(self, filename):
        """Writes the data to the given file::

            DatapointArray([{"t": unix timestamp, "d": data}]).writeJSON("myfile.json")

        The data can later be loaded using loadJSON.
        """
        with open(filename, "w") as f:
            json.dump(self, f)

    def loadJSON(self, filename):
        """Adds the data from a JSON file. The file is expected to be in datapoint format::

            d = DatapointArray().loadJSON("myfile.json")
        """
        with open(filename, "r") as f:
            self.merge(json.load(f))
        return self

    def tshift(self, t):
        """Shifts all timestamps in the datapoint array by the given number of seconds.
        It is the same as the 'tshift' pipescript transform.

        Warning: The shift is performed in-place! This means that it modifies the underlying array::

            d = DatapointArray([{"t":56,"d":1}])
            d.tshift(20)
            print(d) # [{"t":76,"d":1}]
        """
        raw = self.raw()
        for i in range(len(raw)):
            raw[i]["t"] += t
        return self

    def sum(self):
        """Gets the sum of the data portions of all datapoints within"""
        raw = self.raw()
        s = 0
        for i in range(len(raw)):
            s += raw[i]["d"]
        return s

    def mean(self):
        """Gets the mean of the data portions of all datapoints within"""
        return self.sum() / float(len(self))


class Timeseries(Object):
    def __call__(self, actions=False, **kwargs):
        """
        Gets timeseries data. You can query by index with i1 and i2, or by timestamp by t1 and t2.
        Timestamps can be strings such as "last month", "1pm" or "jun 5, 2019, 1pm", which will be
        parsed and converted to the corresponding unix timestamps
        """
        if "t1" in kwargs:
            kwargs["t1"] = parseTime(kwargs["t1"])
        if "t2" in kwargs:
            kwargs["t2"] = parseTime(kwargs["t2"])
        if "t" in kwargs:
            kwargs["t"] = parseTime(kwargs["t"])
        urimod = "/timeseries"
        if actions:
            urimod = "/actions"
        return self.session.get(
            self.uri + urimod, params=kwargs, f=lambda x: DatapointArray(x)
        )

    def __getitem__(self, getrange):
        """Allows accessing the timeseries just as if it were just one big python array.
        An example::

            #Returns the most recent 5 datapoints from the timeseries
            timeseries[-5:]

            #Returns all the data the timeseries holds.
            timeseries[:]

        In order to perform transforms on the timeseries and to aggregate data, look at __call__,
        which allows getting index ranges along with a transform.
        """
        if not isinstance(getrange, slice):
            # Return the single datapoint
            return self(i1=getrange, i2=getrange + 1)[0]

        # The query is a slice - return the range
        qkwargs = {}
        if getrange.start is not None:
            qkwargs["i1"] = getrange.start
        if getrange.stop is not None:
            qkwargs["i2"] = getrange.stop
        return self(**qkwargs)

    def length(self, actions=False):
        urimod = "/timeseries/length"
        if actions:
            urimod = "/actions/length"
        return self.session.get(self.uri + urimod)

    def insert_array(self, datapoint_array, **kwargs):
        """given an array of datapoints, inserts them to the timeseries. This is different from append(),
        because it requires an array of valid datapoints, whereas append only requires the data portion
        of the datapoint, and fills out the rest::

            s.insert_array([{"d": 4, "t": time.time()},{"d": 5, "t": time.time(), "dt": 5.3}])

        Each datapoint can optionally also contain a "dt" parameter with the datapoint's duration in seconds.
        A time series can't have multiple datapoints with the same timestamp, so such datapoints are automatically
        overwritten by default. Using method="insert" will throw an error if a timestamp conflicts with an existing one.
        """
        return self.session.post(
            self.uri + "/timeseries", data=datapoint_array, params=kwargs
        )

    def append(self, data, duration=0):
        """inserts one datapoint with the given data, and appends it to
        the timeseries, using the current timestamp::

            s.append("Hello World!")

        """
        return self.insert_array([{"d": data, "t": time.time(), "dt": duration}])

    def insert(self, data, timestamp=None, duration=0):
        if timestamp is None:
            return self.append(data, duration)
        return self.insert_array([{"t": timestamp, "d": data, "dt": duration}])

    def remove(self, **kwargs):
        """
        Removes the given data from the timeseries
        """
        return self.session.delete(self.uri + "/timeseries", params=kwargs)

    def __len__(self):
        return self.length()


registerObjectType("timeseries", Timeseries)



from typing import Dict
from .base import APIObject, Session
from .kv import KV

from . import apps
from . import objects
from .notifications import Notifications


class User(APIObject):
    props = {"name", "username", "description", "icon"}

    def __init__(self, username: str, session: Session,cached_data={}):
        super().__init__(f"api/users/{username}", {"user": username}, session,cached_data=cached_data)
        self._username = username

        # Apps represents a list of the user's active apps. Apps can be accessed by ID::
        #   myapp = await u.apps["appid"]
        # They can also by queried, which will return a list::
        #   myapp = await u.apps(plugin="myplugin")
        # Finally, they can be created::
        #   myapp = await u.apps.create()
        self.apps = apps.Apps({"owner": username}, self.session)
        self.objects = objects.Objects({"owner": username}, self.session)
        self._kv = KV(f"api/kv/users/{username}", self.session)

    @property
    def kv(self):
        return self._kv
    @property
    def username(self):
        return self._username

    @kv.setter
    def kv(self, v):
        return self._kv.set(**v)


class Users(APIObject):
    def __init__(self, constraints: Dict, session: Session):
        super().__init__("api/users", constraints, session)

    def __getitem__(self, item):
        return self._getitem(item, f=lambda x: User(x["id"], session=self.session,cached_data=x))

    def __call__(self, **kwargs):
        return self._call(
            f=lambda x: [User(xx["id"], session=self.session,cached_data=xx) for xx in x], **kwargs
        )

    def create(self, username, password, **kwargs):
        return self._create(
            f=lambda x: User(x["id"], session=self.session,cached_data=x),
            **{"username": username, "password": password, **kwargs},
        )




from .base import Session
from typing import Dict

class Notifications:
    def __init__(self,constraints : Dict, session : Session):
        self.constraints = constraints
        self.session = session
    def __call__(self,**kwargs):
        return self.session.get("api/notifications",params={**self.constraints,**kwargs})
    def notify(self,key,title=None,**kwargs):
        n = {
            **self.constraints,
            **{'key': key},
            **kwargs
        }
        if title is not None:
            n["title"] = title
        if "_global" in n:
            n["global"] = n["_global"]
            del n["_global"]
        return self.session.post("api/notifications",n)
    def delete(self,key=None,**kwargs):
        if key is not None:
            kwargs["key"] = key
        return self.session.delete("/api/notifications", params={**self.constraints,**kwargs})


from .base import getSessionType
import json
import os
import multidict
import aiohttp
import base64

from .notifications import Notifications

from .users import Users
from .apps import Apps
from .objects import Objects


class Plugin:
    def __init__(self, config=None, session: str = "async"):
        # Load the plugin configuration
        self.config = config
        if self.config is None:
            self.config = json.loads(input())

            # Change the directory to the data dir
            os.chdir(self.config["data_dir"])

        self.session = getSessionType(
            session, self.name, f"http://localhost:{self.config['config']['port']}"
        )
        self.session.setPluginKey(self.config["apikey"])

        self.notifications = Notifications({}, self.session)
        self.apps = Apps({}, self.session)
        self.users = Users({}, self.session)
        self.objects = Objects({}, self.session)

    @property
    def name(self):
        return self.config["plugin"]

    def copy(self):
        return Plugin(self.config, "async" if self.session.isasync else "sync")

    def query_as(self, accessor):
        p = self.copy()
        p.session.setHeader("X-Heedy-As", accessor)
        return p

    async def forward(
        self, request, data=None, headers={}, run_as: str = None, overlay=None
    ):
        """
        Forwards the given request to the underlying database.
        It only functions in async mode.

        Returns the response.
        """
        if data is None:
            data = await request.read()
        headers = {**request.headers, **headers}

        if overlay is None:
            overlay = "next"

        headers["X-Heedy-Overlay"] = overlay

        return await self.session.raw(
            request.method,
            request.path,
            headers=headers,
            data=data,
            params=request.query,
        )

    async def respond_forwarded(self, request, **kwargs):
        """
        Responds to the request with the result of forward()
        """
        req_res = await self.forward(request, **kwargs)

        response = aiohttp.web.TimeseriesResponse(
            status=req_res.status, headers=req_res.headers
        )
        await response.prepare(request)
        while True:
            chunk = await req_res.content.read(32768)
            if not chunk:
                break
            await response.write(chunk)
        await response.write_eof()
        return response

    def objectRequest(self, request):
        h = request.headers
        last_modified = h["X-Heedy-Last-Modified"]
        if last_modified == "null":
            last_modified = None
        return {
            "request": h["X-Heedy-Request"],
            "id": h["X-Heedy-Id"],
            "last_modified": last_modified,
            "meta": json.loads(base64.b64decode(h["X-Heedy-Meta"])),
            "object": h["X-Heedy-Object"],
            "type": h["X-Heedy-Type"],
            "owner": h["X-Heedy-Owner"],
            "as": h["X-Heedy-As"],
            "access": h["X-Heedy-Access"].split(" "),
        }

    def hasAccess(self, request, scope):
        a = request.headers["X-Heedy-Access"].split(" ")
        return "*" in a or scope in a

    def isUser(self, request):
        h = request.headers
        return (
            not "/" in h["X-Heedy-As"]
            and h["X-Heedy-As"] != "heedy"
            and h["X-Heedy-As"] != "public"
        )

    def isApp(self, request):
        return "/" in request.headers["X-Heedy-As"]

    def isAdmin(self, request):
        request.headers["X-Heedy-As"] == "heedy"

    def fire(self, event):
        """
        Fires the given event
        """
        return self.session.post("/api/events", event)

    def notify(self, *args, **kwargs):
        return self.notifications.notify(*args, **kwargs)



import pytest
from heedy import App


def test_basics():
    a = App("testkey")
    assert a.owner.username == "test"

    a.owner.name = "Myname"
    assert a.owner.name == "Myname"

    assert len(a.objects()) == 0

    o = a.objects.create("myobj", {"schema": {"type": "number"}})
    assert o.name == "myobj"
    assert o.type == "timeseries"
    assert len(a.objects()) == 1

    assert o == a.objects[o.id]

    assert o.length() == 0
    o.append(2)
    assert o.length() == 1
    d = o[:]
    assert len(d) == 1
    assert d[0]["d"] == 2
    assert "dt" not in d[0]
    o.append(3, duration=9)
    d = o[:]
    assert len(d) == 2
    assert d[1]["d"] == 3
    assert d[1]["dt"] == 9
    o.remove()  # Clear the timeseries
    assert len(o) == 0

    o.delete()
    assert len(a.objects()) == 0
    # assert len(a.owner.apps())==1


def test_notifications():
    a = App("testkey")

    assert len(a.notifications()) == 0
    a.notify("hi", "hello")
    assert len(a.notifications()) == 1
    a.notifications.delete("hi")
    assert len(a.notifications()) == 0


def test_kv():
    a = App("testkey")

    assert len(a.kv()) == 0
    a.kv["test"] = True
    assert len(a.kv()) == 1
    assert a.kv["test"] == True

    del a.kv["test"]
    assert len(a.kv()) == 0


@pytest.mark.asyncio
async def test_basics_async():
    a = App("testkey", session="async")

    await (await a.owner).update(name="Myname2")
    assert await (await a.owner).name == "Myname2"

    assert len(await a.objects()) == 0

    o = await a.objects.create("myobj2", {"schema": {"type": "number"}})
    assert o.name == "myobj2"
    assert o.type == "timeseries"
    assert len(await a.objects()) == 1

    assert o == await a.objects[o.id]

    await o.delete()

    # assert len(await (await a.owner).apps())==1



import pathlib
import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setuptools.setup(
    name="heedy",
    version="0.0.8",
    description="A Python library for interfacing with Heedy",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/heedy/heedy",
    author="Heedy Contributors",
    license="apache2",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires=">=3.7.0",
    install_requires=["numpy", "aiohttp", "dateparser", "requests"],
)



from typing import Dict
from ..base import Session

from . import objects

# This map contains the registered object types
_objectRegistry = {}


def registerObjectType(objectType: str, objectClass) -> None:
    """
    registerObjectType allows external libraries to implement object types
    available through heedy plugins. All you need to do is subclass :code:`Object`,
    and register the corresponding type!
    """
    _objectRegistry[objectType] = objectClass


def getObject(objectData: Dict, session: Session):
    """
    Heedy allows multiple different object types. getObject uses the
    registered object type to initialize the given data to
    the correct class. If the object is of an unregistered type, it returns
    a base :code:`Object` object.
    """
    if objectData["type"] in _objectRegistry:
        return _objectRegistry[objectData["type"]](objectData,session)
    return objects.Object(objectData,session)
